<?php


namespace App\Http\Controllers\Patch\Liason;


use App\Exceptions\GeneralException;
use App\Filters\Patch\Agent\AgentFilter;
use App\Filters\Patch\Customer\ClientFilter;
use App\Filters\Patch\Invoice\InvoiceFilter;
use App\Filters\Patch\Reports\PaymentSummaryFilter;
use App\Filters\Patch\Support\AgentUserFilter;
use App\Filters\Patch\Support\ClientUserFilter;
use App\Helpers\traits\DateRangeHelper;
use App\Http\Controllers\Controller;
use App\Models\Core\Status;
use App\Models\Currency;
use App\Models\Invoice;
use App\Models\App\PaymentMethods\PaymentMethod;

use App\Models\Core\Auth\User;
use App\Models\Leads;
use App\Models\Patch\Recurring\RecurringCycle;
use App\Models\Patch\Tax\Tax;
use App\Models\Payment;
use App\Models\Property;
use App\Models\Unit;
use App\Repositories\Core\Status\StatusRepository;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\DB;

use App\Models\Core\Auth\Role;
use PayPal\Api\PaymentHistory;
use Illuminate\Database\Eloquent\Builder;


class SupportApiController  extends Controller
{


    use DateRangeHelper;


    public function invoice()
    {
        $paidInvoiceStatus = resolve(StatusRepository::class)->invoicePaid();
        return Invoice::query()
            ->filters(new InvoiceFilter())
            ->where('status_id', '<>', $paidInvoiceStatus)
            ->select('id', 'invoice_number', 'due_amount')
            ->orderByDesc('id')
            ->paginate(request('per_page', 10));
    }
    public function invoiceShow(Invoice $invoice)
    {
        return $invoice;
    }

    public function recurringCycle()
    {
        return RecurringCycle::select('id', 'name')->get();
    }


    public function paymentMethod()
    {
        $activePaymentMethod = Status::findByNameAndType('status_active', 'payment_method')->id;

        return PaymentMethod::query()
            ->where('status_id', '=', $activePaymentMethod)
            ->select('id', 'name', 'alias')
            ->get();
    }

    public function agentUser()
    {
        return User::query()
            ->whereHas('roles', function ($query) {
            $query->where('alias', '=', 'agent');
        })
            ->filters(new AgentUserFilter())
            ->select('id', 'first_name', 'last_name')
            ->orderByDesc('id')
            ->paginate(request('per_page', 10));
    }
    public function clientUser()
    {
        return User::query()->whereHas('roles', function ($query) {
            $query->where('alias', '=', 'client');
        })
            ->filters(new ClientUserFilter())
            ->select('id', 'first_name', 'last_name')
            ->orderByDesc('id')
            ->paginate(request('per_page', 10));
    }


    public function currency()
    {
        return Currency::select('id', 'name')->get();
    }

    public function invoiceNumber()
    {
        return Invoice::withTrashed()->select('id')->orderByDesc('id')->first();
    }


    public function taxes()
    {
        return Tax::select('id', 'name', 'value')->get();
    }
    public function roles()
    {
        return Role::with('users:id,first_name,last_name,email', 'users.profilePicture')
            ->whereNull('alias')
            ->orderBy('id')
            ->get();
    }

    public function invoiceSummation()
    {
        return Invoice::select(
            DB::raw('SUM(total) as total_amount'),
            DB::raw('SUM(received_amount) as paid_amount'),
            DB::raw('SUM(due_amount) as due_amount'),
            DB::raw('SUM(processing_fee) as processing_fee')
        )->first();
    }

    public function paymentSummation(): array
    {
        return [
            'total_amount' => $this->paymentSummationQuery(),
            'total_this_month_amount' => $this->paymentSummationQuery($this->getStartAndEndOf('thisMonth')),
            'total_last_month_amount' => $this->paymentSummationQuery($this->getStartAndEndOf('lastMonth')),
            'total_this_year_amount' => $this->paymentSummationQuery($this->getStartAndEndOf('thisYear')),
        ];
    }

    private function paymentSummationQuery($dateRange = [])
    {
        return Payment::query()
            ->when(count($dateRange), function (Builder $query) use ($dateRange) {
                $query->whereBetween('received_on', $dateRange);
            })->where('payment_status_id',15)
            ->filters(new PaymentSummaryFilter())->sum('amount');
    }

    public function property()
    {
        return Property::query()->select('id', 'property_name', 'property_code')->get();
    }

    public function unit()
    {
        return Unit::query()->select('id', 'unit_name', 'sale_amount')->get();
    }


    public function leads()
    {
        return Leads::with('customer:id,full_name', 'agent:id,first_name', 'unit:id,unit_name,sale_amount,unit_floor,total_rooms,square_foot,status_id','property:id,project_name')
            ->orderBy('id')
            ->get();
    }

    public function stopRecurring(Invoice $invoice)
    {
        throw_if(!auth()->user()->isAppAdmin(), new GeneralException(trans('default.403')));

        $invoice->update([
            'recurring' => 2,
            'recurring_cycle_id' => null
        ]);
        return updated_responses('stop_recurring');

    }

    public function cacheClear()
    {
        throw_if(!auth()->user()->isAppAdmin(), new GeneralException(trans('default.403')));

        Artisan::call('optimize:clear');

        return response()->json([
            'message' => __t('cache_clear_successfully')
        ]);
    }






}
