<?php

namespace App\Services\App\Dashboard;


use App\Filters\Patch\Dashboard\DashboardFilter;
use App\Models\Core\Status;
use App\Models\Invoice;
use App\Models\Leads;
use App\Models\Payment;
use App\Models\Property;
use App\Models\Prospect;
use App\Models\ReservedUnitPayment;
use App\Models\Unit;
use App\Models\UnitBooking;
use App\Services\Patch\ApplicationBaseService;
use Carbon\Carbon;
use Carbon\CarbonPeriod;
use App\Models\Core\Auth\User;
use Illuminate\Database\Query\Builder;
use Illuminate\Support\Facades\DB;
use stdClass;
use Illuminate\Support\Facades\Auth;


class DashboardService extends ApplicationBaseService
{
    public $status;
    public $documentStatus;
    public DashboardFilter $filter;
    protected Payment $paymentModel;

    public function __construct(Invoice $invoice, DashboardFilter $filter, Payment $paymentModel)
    {
        $this->model = $invoice;
        $this->filter = $filter;
        $this->status = $this->getStatus();
        $this->documentStatus = $this->getStatus('document');
        $this->paymentModel = $paymentModel;
    }
    public function totalAgent()
    {
        return User::whereHas('roles', function ($query) {
            $query->where('alias', 'agent');
        })->where('status_id', $this->status['status_active'])
            ->count();
    }

    public function totalReservedLeeds($propertyIds = null)
    {
        $query = ReservedUnitPayment::where('status_id', 12);

        // Check if the user is not an admin
        if (!Auth::user()->isAppAdmin()) {
            // If not admin, filter by created_by user ID
            $query->where('created_by', Auth::id());
        }

        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }

        // Return the count of the filtered query
        return $query->count();
    }

    public function totalProspects($propertyIds = null)
    {
        $query = Prospect::where('status_id', 18);

        // Check if the user is not an admin
        if (!Auth::user()->isAppAdmin()) {
            // If not admin, filter by created_by user ID
            $query->where('user_id', Auth::id());
        }

        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }

        // Return the count of the filtered query
        return $query->count();
    }


    public function offerLetter($propertyIds = null, $status = null)
    {
        $query = UnitBooking::whereHas('unit', function ($q) {
            // Ensure amount in UnitBooking is at least 25% of the sale_amount in Unit
            $q->whereRaw('amount < 0.25 * sale_amount');
        });

//        if ($status){
//            $query->when($status, function ($query) use ($status) {
//                return $query->where('status_id', $status->id);
//            });
//        }

        if (!Auth::user()->isAppAdmin()) {
            $query->where('user_id', Auth::id());
        }

        if (!empty($propertyIds)){
            $query ->whereIn('property_id', $propertyIds);
        }

        return $query->count();
    }



    public function saleAgreement($propertyIds = null)
    {
        $query = UnitBooking::whereHas('unit', function ($q) {
            // Ensure amount in UnitBooking is at least 25% of the sale_amount in Unit
            $q->whereRaw('amount >= 0.25 * sale_amount');
        });

        if (!Auth::user()->isAppAdmin()) {
            $query->where('user_id', Auth::id());
        }

        if (!empty($propertyIds)){
            $query ->whereIn('property_id', $propertyIds);
        }
        return $query->count();
    }

    public function invoiceCount($status = null, $propertyIds= null)
    {
        $query = $this->model->query()
            ->filters($this->filter)
            ->when($status, function ($query) use ($status) {
                return $query->where('status_id', $status);
            });

        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }
        return $query->count();
    }

    public function totalClient()
    {
        return User::whereHas('roles', function ($query) {
            $query->where('alias', 'client');
        })->where('status_id', $this->status['status_active'])
            ->count();
    }


    public function totalProperty($propertyIds = null): int
    {
        $query = Leads::query();
        if ($propertyIds){
            $query->whereIn('property_id', $propertyIds);
        }
        return $query->count();
    }
    public function totalUnits($propertyIds = null)
    {
        $query = Unit::query();
        if (!empty($propertyIds)){
            $query->whereIn('property_id', $propertyIds);
        }
        return $query->count();
    }

    public function totalUnitsBooked($propertyIds = null)
    {
        $query = UnitBooking::query();
        if (!empty($propertyIds)){
            $query->whereIn('property_id', $propertyIds);
        }
        return$query->count();
    }

    public function totalProperties($propertyIds = null)
    {
        $query = Property::query();
        if (!empty($propertyIds)){
            $query->whereIn('id', $propertyIds);
        }
        return$query->count();
    }


    public function paymentSummation($columnName, $propertyIds)
    {
        $query = $this->model->query();

        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }

        return $query->sum($columnName);
    }


    public function paymentOverView($propertyIds = null)
    {
        $query =  $this->model->query()
            ->selectRaw('SUM(received_amount) as received_amount, SUM(due_amount) as due_amount');
        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }
        return $query->get();
//        return $this->model
//            ->query()
//            ->filters($this->filter)
//            ->select(
//                DB::raw('SUM(received_amount) as received_amount'),
//                DB::raw('SUM(due_amount) as due_amount')
//            )->get();
    }

    public function monthlyIncome($propertyIds = null)
    {
        $query = $this->model
            ->query()
            ->whereBetween('date', $this->thisYear())
            ->selectRaw('SUM(invoices.received_amount) as income, MONTH(invoices.date) as month')
            ->groupBy('month')
            ->orderBy('month');

        if (!empty($propertyIds)) {
            $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereIn('property_id', $propertyIds);
            });
        }

        return $query->get();
    }

    public function monthlyIncomeByPayments($propertyIds = null)
    {
        $query = $this->paymentModel
            ->query()
            ->whereBetween('received_on', $this->thisYear())
            ->selectRaw('SUM(payments.amount) as income, MONTH(payments.received_on) as month')
            ->groupBy('month')
            ->orderBy('month');

        if (!empty($propertyIds)) {
            $query->whereHas('invoice', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                $query->whereHas('unit', function (\Illuminate\Database\Eloquent\Builder $query) use ($propertyIds) {
                    $query->whereIn('property_id', $propertyIds);
                });
            });
        }

        return $query->get();
    }

    private function thisYear(): array
    {
        return array(now()->startOfYear()->toDateTimeString(), now()->endOfYear()->toDateTimeString());
    }

    public function manipulateChart(object $charts): array
    {
        $monthContainer = [];
        for ($i = 1; $i <= 12; $i++) {
            $monthContainer[Carbon::create()->month($i)->format('M')] = new stdClass();
        }
        foreach ($charts as $value) {
            $monthContainer[Carbon::create()->month($value->month)->format('M')] = $value;
        }
        return $monthContainer;
    }

    public function getStatus($type = 'invoice')
    {
        return Status::where('type', 'user')
            ->orWhere('type', $type)
            ->pluck('id', 'name');
    }


}
